/*
** OpenAmiga.org
**  This file is part of the OpenAmiga "XAD filesystem"
**
** For more information look at:
**   http://openamiga.org/?function=viewproject&projectid=66
*/

#include "xadfs.h"
#include "XADFileSystem_rev.h"

static CONST TEXT USED verstag[] = VERSTAG;

static int32 FSUADummy(struct FSVP *vp, int32 *res2) {
	*res2 = ERROR_ACTION_NOT_KNOWN;
	return DOSFALSE;
}

static int32 FSWPDummy(struct FSVP *vp, int32 *res2) {
	*res2 = ERROR_WRITE_PROTECTED;
	return DOSFALSE;
}

static CONST struct FileSystemVectors FSVectors = {
	sizeof (struct FileSystemVectors), /* StructSize */
	FS_VECTORPORT_VERSION,             /* Version  */
	0,                /* FSPrivate - We use this for globaldata */
	{ 0, 0, 0 },      /* Reserved[3] */
	0,                /* DOSPrivate  */
	NULL,             /* DOS will set this to DOSEmulatePacket() */

	FSLock,
	FSUnLock,
	FSDupLock,
	(APTR)FSWPDummy, /* FSCreateDir */
	FSParentDir,
	FSDupLockFromFH,
	FSOpenFromLock,
	FSParentOfFH,
	FSOpen,
	FSClose,
	(APTR)FSWPDummy, /* FSDelete */
	FSRead,
	FSWrite,
	FSFlush,

	FSChangeFilePosition,
	(APTR)FSWPDummy, /* FSChangeFileSize */

	FSGetFilePosition,
	FSGetFileSize,

	FSChangeLockMode,
	FSChangeFileMode,

	(APTR)FSWPDummy, /* FSSetDate */
	(APTR)FSWPDummy, /* FSSetProtection */
	(APTR)FSWPDummy, /* FSSetComment */
	(APTR)FSWPDummy, /* FSSetGroup */
	(APTR)FSWPDummy, /* FSSetOwner */
	(APTR)FSWPDummy, /* FSRename */

	(APTR)FSWPDummy, /* FSCreateSoftLink */
	(APTR)FSWPDummy, /* FSCreateHardLink */
	FSReadSoftLink,

	FSSameLock,
	FSSameFile,
	FSFileSystemAttr,
	FSVolumeInfoData,
	FSDeviceInfoData,

	NULL, /* reserved */
	NULL, /* reserved */

	FSExamineLock,
	FSExamineFile,
	FSExamineDir,
	(APTR)FSUADummy, /* FSInhibit */
	(APTR)FSWPDummy, /* FSWriteProtect */
	(APTR)FSWPDummy, /* FSFormat */
	(APTR)FSWPDummy, /* FSSerialize */
	(APTR)FSWPDummy, /* FSRelabel */

	NULL, /* reserved */

	(APTR)FSUADummy, /* FSAddNotify */
	(APTR)FSUADummy, /* FSRemoveNotify */

	(APTR)FSUADummy, /* FSLockRecord */
	(APTR)FSUADummy, /* FSUnLockRecord */

	-1 /* V53_End_Marker */
};

static void handle_action_shutdown(struct GlobalData *gd, struct DosPacket *dp) {
	struct ObjLock *lock, *next;
	struct MsgPort *port;

	if (gd->debug_level > 5) {
		IEXEC->DebugPrintF("[FS] ACTION_SHUTDOWN packet received.\n");
	}

	if (NO gd->FSVectorPort->FSV.Version) {
		dp->dp_Res1 = DOSFALSE;
		dp->dp_Res2 = ERROR_ACTION_NOT_KNOWN;
		return;
	}

	IEXEC->ObtainSemaphore(gd->Sem);

	gd->FSVectorPort->FSV.Version = 0;

	if (gd->volume_node) {
		gd->volume_node->vn_DiskType = ID_NO_DISK_PRESENT;
		IDOS->NonBlockingModifyDosEntry(NULL, NBM_DISKREMOVED, NULL, NULL);
	}

	if (gd->device_node) {
		gd->device_node->dn_Port = NULL;
		gd->device_node = NULL;
	}

	if ((port = IEXEC->FindPort("dos_lock_handler_port"))) {
		lock = BADDR(gd->lock_list);
		while (lock) {
			next = BADDR(lock->lock.fl_Link);

			if (lock->fh) {
				IDOS->DoPkt(port,ACTION_COLLECT, ID_COLLECT_FILEHANDLE, (int32)lock->fh, 0, 0, 0);

				IDOS->FreeDosObject(DOS_LOCK, lock);
			} else {
				IDOS->DoPkt(port, ACTION_COLLECT, ID_COLLECT_LOCK, (int32)lock, 0, 0, 0);
			}

			lock = next;
		}
	}

	IEXEC->ReleaseSemaphore(gd->Sem);

	if (gd->debug_level > 5) {
		IEXEC->DebugPrintF("[FS] Dismounting completed.\n");
	}

	dp->dp_Res1 = DOSTRUE;
	dp->dp_Res2 = 0;
}

static int32 filesystem(struct ExecBase *sysbase) {
	struct ExecIFace *iexec = (struct ExecIFace *)sysbase->MainInterface;
	struct DosPacket *startup_packet;
	int32 shutdown = FALSE;
	struct GlobalData *gd;
	struct Process *thisproc;
	struct DosPacket *dp;
	struct Message *msg;
	struct MsgPort *msgport;
	TEXT devName[256];
	TEXT archiveName[256];

	thisproc = (struct Process *)iexec->FindTask(NULL);
	msgport = &thisproc->pr_MsgPort;

	iexec->WaitPort(msgport);
	msg = iexec->GetMsg(msgport);
	startup_packet = (struct DosPacket *)msg->mn_Node.ln_Name;
	if (NO startup_packet) {
		iexec->ReplyMsg(msg);
		iexec->DebugPrintF("[FS] Ran from WB, Eeek!\n");
		return RETURN_FAIL;
	}

	gd = iexec->AllocVecTags(sizeof(*gd),
		AVT_Type, MEMF_SHARED,
		AVT_Lock, TRUE,
		AVT_ClearWithValue, 0,
		TAG_END);
	if (NO gd) {
		startup_packet->dp_Res1 = DOSFALSE;
		startup_packet->dp_Res2 = ERROR_NO_FREE_STORE;
		iexec->ReplyMsg(msg);
		iexec->DebugPrintF("[FS] No memory for GD\n");
		return RETURN_FAIL;
	}

	gd->SysBase = (APTR)sysbase;
	gd->IExec   = iexec;

	gd->DOSBase = IEXEC->OpenLibrary("dos.library", 53);
	if (NO gd->DOSBase) {
		IEXEC->DebugPrintF("[FS] Can't open V53 dos.library\n");
		goto cleanup;
	}

	gd->UtilityBase = IEXEC->OpenLibrary("utility.library", 53);
	if (NO gd->UtilityBase) {
		IEXEC->DebugPrintF("[FS] Can't open V53 utility.library\n");
		goto cleanup;
	}

	gd->XADBase = IEXEC->OpenLibrary("xadmaster.library", 13);
	if (NO gd->XADBase) {
		IEXEC->DebugPrintF("[FS] Can't open V13 xadmaster.library\n");
		goto cleanup;
	}

	gd->IDos = (struct DOSIFace *)IEXEC->GetInterface(gd->DOSBase, "main", 1, NULL);
	if (NO gd->IDos) {
		goto cleanup;
	}

	gd->IUtility = (struct UtilityIFace *)IEXEC->GetInterface(gd->UtilityBase, "main", 1, NULL);
	if (NO gd->IUtility) {
		goto cleanup;
	}

	gd->IXad = (struct XadMasterIFace *)IEXEC->GetInterface(gd->XADBase, "main", 1, NULL);
	if (NO gd->IXad) {
		goto cleanup;
	}

	gd->device_node = BADDR(startup_packet->dp_Arg3);
	IDOS->CopyStringBSTRToC(gd->device_node->dn_Name, devName, sizeof(devName));
	gd->device_name = devName;

	if (NOT LIB_IS_AT_LEAST(gd->DOSBase, 53, 95)) {
		static CONST TEXT youneed[] = "[%s] You need dos.library 53.95+\nYou only have V%ld.%ld\n\n";

		IEXEC->DebugPrintF(youneed, devName, gd->DOSBase->lib_Version, gd->DOSBase->lib_Revision);

		IDOS->TimedDosRequesterTags(TDR_Timeout, 30,
			TDR_NonBlocking,  TRUE,
			TDR_TitleString,  VERS,
			TDR_FormatString, youneed,
			TDR_GadgetString, "OK",
			TDR_ImageType,    TDRIMAGE_ERROR,
			TDR_Arg1,         devName, 
			TDR_Arg2,         gd->DOSBase->lib_Version,
			TDR_Arg3,         gd->DOSBase->lib_Revision,
			TAG_END);

		goto cleanup;
	}

	gd->FSVectorPort = IDOS->AllocDosObjectTags(DOS_FSVECTORPORT,
		ADO_Name,    devName,
		ADO_Vectors, &FSVectors,
		TAG_END);
	if (NO gd->FSVectorPort) {
		IEXEC->DebugPrintF("[%s] Cannot allocate vector port.\n", devName);
		goto cleanup;
	}

	gd->FSVectorPort->FSV.FSPrivate = gd;

	gd->Sem = IEXEC->AllocSysObject(ASOT_SEMAPHORE, NULL);
	if (NO gd->Sem) {
		IEXEC->DebugPrintF("[%s] Cannot allocate semaphore.\n", devName);
		goto cleanup;
	}

	gd->memory_pool = IEXEC->AllocSysObjectTags(ASOT_MEMPOOL,
		ASOPOOL_Name,      "XADFileSystem memory pool",
		ASOPOOL_Protected, TRUE,
		ASOPOOL_MFlags,    MEMF_SHARED,
		ASOPOOL_LockMem,   FALSE,
		ASOPOOL_Puddle,    4096,
		ASOPOOL_Threshold, 1024,
		TAG_END);
	if (NO gd->memory_pool) {
		IEXEC->DebugPrintF("[%s] Cannot allocate pool.\n", devName);
		goto cleanup;
	}

	IDOS->CopyStringBSTRToC(startup_packet->dp_Arg2, archiveName, sizeof(archiveName));

	if (NOT mount_xad_archive(gd, archiveName)) {
		IEXEC->DebugPrintF("[%s] Failed to mount archive file '%s'.\n", devName, archiveName);
		goto cleanup;
	}

	gd->volume_node = IDOS->AllocDosObjectTags(DOS_DOSLIST,
		ADO_Type, DLT_VOLUME,
		ADO_Name, IDOS->FilePart(archiveName),
		TAG_END);
	if (NO gd->volume_node) {
		IEXEC->DebugPrintF("[%s] Cannot allocate volume node.\n", devName);
		goto cleanup;
	}

	gd->volume_node->vn_DiskType   = ID_XAD0_DISK;
	gd->volume_node->vn_Port       = &gd->FSVectorPort->MP;
	gd->volume_node->vn_VolumeDate = gd->root_node->date;
	gd->volume_node->vn_FSPrivate  = gd->device_node->dn_Name; /* needed by MountXAD */

	if (NOT IDOS->NonBlockingModifyDosEntry((struct DosList *)gd->volume_node, NBM_ADDDOSENTRY, NULL, NULL)) {
		IEXEC->DebugPrintF("[%s] Cannot add volume node to doslist.\n", devName);
		IDOS->FreeDosObject(DOS_DOSLIST, gd->volume_node);
		gd->volume_node = NULL;
		goto cleanup;
	}

	gd->device_node->dn_Port = &gd->FSVectorPort->MP;
	IDOS->ReplyPkt(startup_packet, DOSTRUE, 0);
	startup_packet = NULL;

	if (gd->debug_level > 5) {
		IEXEC->DebugPrintF("[%s] Handler has started successfully.\n", devName);
	}

	while (NOT shutdown) {
		IEXEC->WaitPort(&gd->FSVectorPort->MP);

		while ((msg = IEXEC->GetMsg(&gd->FSVectorPort->MP))) {
			if ((dp = (struct DosPacket *)msg->mn_Node.ln_Name)) {
				if (dp->dp_Link == msg) {
					if (dp->dp_Action == ACTION_SHUTDOWN) {
						handle_action_shutdown(gd, dp);
						shutdown = TRUE;
					} else {
						gd->FSVectorPort->FSV.DOSEmulatePacket(gd->FSVectorPort, dp);
					}
				}
			}

			IEXEC->ReplyMsg(msg);
		}
	}

cleanup:
	if (gd->debug_level > 5) {
		IEXEC->DebugPrintF("[FS] Handler cleanup in progress...\n");
	}

	if (NOT shutdown) {
		if (gd->FSVectorPort) {
			gd->FSVectorPort->FSV.Version = 0;
		}
		if (gd->device_node) {
			gd->device_node->dn_Port = NULL;
			gd->device_node = NULL;
		}
	}

	if (gd->IXad) {
		unmount_xad_archive(gd);
	}

	IEXEC->FreeSysObject(ASOT_MEMPOOL, gd->memory_pool);
	IEXEC->FreeSysObject(ASOT_SEMAPHORE, gd->Sem);
	IEXEC->DropInterface((struct Interface *)gd->IXad);
	IEXEC->CloseLibrary(gd->XADBase);
	IEXEC->DropInterface((struct Interface *)gd->IUtility);
	IEXEC->CloseLibrary(gd->UtilityBase);

	if (gd->IDos) {
		IDOS->NonBlockingModifyDosEntry((struct DosList *)gd->volume_node, NBM_REMFREEDOSENTRY, NULL, NULL);
		IDOS->FreeDosObject(DOS_FSVECTORPORT, gd->FSVectorPort);

		IDOS->ReplyPkt(startup_packet, FALSE, IDOS->IoErr());
		IEXEC->DropInterface((struct Interface *)gd->IDos);
	}

	IEXEC->DropInterface((struct Interface *)gd->IDos);
	IEXEC->CloseLibrary(gd->DOSBase);

	if (gd->debug_level > 5) {
		IEXEC->DebugPrintF("[FS] Handler cleanup completed, process exiting...\n");
	}

	IEXEC->UnlockMem(gd, sizeof(*gd));
	IEXEC->FreeVec(gd);

	return RETURN_OK;
}

static int32 cli_startup(struct ExecBase *sysbase) {
	struct ExecIFace *iexec = (struct ExecIFace *)sysbase->MainInterface;
	struct Library *dosbase;
	struct DOSIFace *idos;
	int32 result = RETURN_ERROR;

	if ((dosbase = iexec->OpenLibrary("dos.library", 53)) != NULL) {
		if ((idos = (struct DOSIFace *)iexec->GetInterface(dosbase, "main", 1, NULL)) != NULL) {
			result = RETURN_OK;

			idos->Printf("\n%s [Vectorport Filesystem]\n", VERS);

			iexec->DropInterface((struct Interface *)idos);
		}
		iexec->CloseLibrary(dosbase);
	}

	return result;
}

int32 _start(STRPTR argstring, int32 arglen, struct ExecBase *sysbase) {
	struct ExecIFace *iexec = (struct ExecIFace *)sysbase->MainInterface;
	struct Process *thisproc = (struct Process *)iexec->FindTask(NULL);
	int32 result;

	iexec->Obtain();

	if (thisproc->pr_CLI)
		result = cli_startup(sysbase);
	else
		result = filesystem(sysbase);

	iexec->Release();

	return result;
}

