/*
** OpenAmiga.org
**  This file is part of the OpenAmiga "XAD filesystem"
**
** For more information look at:
**   http://openamiga.org/?function=viewproject&projectid=66
*/

#include "xadfs.h"

static int32 generate_xad_dir_tree(struct GlobalData *gd);
static void free_xad_dir_tree(struct GlobalData *gd);
static struct ObjNode *alloc_obj_node(struct GlobalData *gd, CONST_STRPTR name,
	struct ObjNode *parent_dir);
static void free_obj_node(struct GlobalData *gd, struct ObjNode *obj);
static int32 create_cache_directory(struct GlobalData *gd);
static void delete_cache_directory(struct GlobalData *gd);
static void delete_cache_directory_contents(struct GlobalData *gd, BPTR dirLock);
static int32 extract_xad_file(struct GlobalData *gd, int32 *res2, struct ObjNode *obj);

static int32 xad_to_ioerror(int32 xad_error) {
	switch (xad_error) {
	case XADERR_OK:           return 0;
	case XADERR_UNKNOWN:      return -1;
	case XADERR_INPUT:        return ERROR_BUFFER_OVERFLOW;
	case XADERR_OUTPUT:       return ERROR_BUFFER_OVERFLOW;
	case XADERR_BADPARAMS:    return -1;
	case XADERR_NOMEMORY:     return ERROR_NO_FREE_STORE;
	case XADERR_ILLEGALDATA:  return -1;
	case XADERR_NOTSUPPORTED: return ERROR_NOT_IMPLEMENTED;
	case XADERR_RESOURCE:     return -1;
	case XADERR_DECRUNCH:     return -1;
	case XADERR_FILETYPE:     return -1;
	case XADERR_OPENFILE:     return -1;
	case XADERR_SKIP:         return -1;
	case XADERR_BREAK:        return ERROR_BREAK;
	case XADERR_FILEEXISTS:   return -1;
	case XADERR_PASSWORD:     return -1;
	case XADERR_MAKEDIR:      return -1;
	case XADERR_CHECKSUM:     return -1;
	case XADERR_VERIFY:       return -1;
	case XADERR_GEOMETRY:     return -1;
	case XADERR_DATAFORMAT:   return -1;
	case XADERR_EMPTY:        return -1;
	case XADERR_FILESYSTEM:   return -1;
	case XADERR_FILEDIR:      return -1;
	case XADERR_SHORTBUFFER:  return -1;
	case XADERR_ENCODING:     return -1;
	default:                  return -1;
	}
}

int32 mount_xad_archive(struct GlobalData *gd, CONST_STRPTR archiveName) {
	struct ExamineData *exd;
	int32 xad_error;

	gd->archive_file = IDOS->Open(archiveName, MODE_OLDFILE);
	if (NOT gd->archive_file) {
		return DOSFALSE;
	}

	gd->xad_archive_info = IXAD->xadAllocObject(XADOBJ_ARCHIVEINFO, NULL);
	if (NOT gd->xad_archive_info) {
		IDOS->SetIoErr(ERROR_NO_FREE_STORE);
		return DOSFALSE;
	}

	xad_error = IXAD->xadGetInfo(gd->xad_archive_info,
		XAD_INFILEHANDLE, gd->archive_file,
		TAG_END);
	if (xad_error) {
		IDOS->SetIoErr(xad_to_ioerror(xad_error));
		return DOSFALSE;
	}

	if (NOT generate_xad_dir_tree(gd)) {
		return DOSFALSE;
	}

	exd = IDOS->ExamineObjectTags(
		EX_FileHandleInput, gd->archive_file,
		TAG_END);
	if (NO exd) {
		return DOSFALSE;
	}

	gd->root_node->date = exd->Date;

	IDOS->FreeDosObject(DOS_EXAMINEDATA, exd);

	return DOSTRUE;
}

void unmount_xad_archive(struct GlobalData *gd) {
	delete_cache_directory(gd);

	free_xad_dir_tree(gd);

	IXAD->xadFreeObject(gd->xad_archive_info, NULL);
	gd->xad_archive_info = NULL;

	IDOS->Close(gd->archive_file);
	gd->archive_file = ZERO;
}

static int32 avl_node_compare_func (struct AVLNode *node1, struct AVLNode *node2) {
	struct ObjNode *obj1 = (struct ObjNode *)node1;
	struct ObjNode *obj2 = (struct ObjNode *)node2;
	struct GlobalData *gd = obj1->global_data;

	return IUTILITY->Stricmp(obj1->name, obj2->name);
}

static int32 avl_key_compare_func (struct AVLNode *node, AVLKey key) {
	struct ObjNode *obj = (struct ObjNode *)node;
	struct GlobalData *gd = obj->global_data;

	return IUTILITY->Stricmp(obj->name, (CONST_STRPTR)key);
}

struct ObjNode *handle_rel_args(struct GlobalData *gd, int32 *res2,
	struct Lock *rel_lock, CONST_STRPTR name)
{
	struct ObjLock *rel = (struct ObjLock *)rel_lock;
	struct ObjNode *reference;
	struct ObjNode *object = NULL;

	if (rel) {
		reference = rel->node;
	} else {
		reference = gd->root_node;
	}

	if (name[0]) {
		if (reference->type != FSO_TYPE_DIRECTORY) {
			*res2 = ERROR_OBJECT_WRONG_TYPE; /* only directories can have contents */
			return NULL;
		}
		object = (struct ObjNode *)IEXEC->AVL_FindNode(reference->content_list, (AVLKey)name, avl_key_compare_func);
		if (NO object) {
			*res2 = ERROR_OBJECT_NOT_FOUND;
			return NULL;
		}
	} else {
		object = reference;
	}

	return object;
}

static int32 generate_xad_dir_tree(struct GlobalData *gd) {
	struct xadFileInfo *xad_file_info;
	struct ObjNode *obj, *parent_dir;
	CONST_STRPTR path;
	int32 path_pos;
	TEXT name[MAX_FILENAME + 1];

	gd->root_node = alloc_obj_node(gd, NULL, NULL);
	if (NO gd->root_node) {
		return DOSFALSE;
	}

	xad_file_info = gd->xad_archive_info->xai_FileInfo;
	while (xad_file_info) {
		obj = gd->root_node;

		path = xad_file_info->xfi_FileName;
		path_pos = 0;

		if (path[0] == '\0') {
			xad_file_info = xad_file_info->xfi_Next;
			continue;
		}

		do {
			path_pos = IDOS->SplitName(path, '/', name, path_pos, sizeof(name));

			if (name[0] == '\0') {
				continue;
			}

			if (obj->type != FSO_TYPE_DIRECTORY) {
				IDOS->SetIoErr(ERROR_OBJECT_WRONG_TYPE); /* only directories can have contents */
				return DOSFALSE;
			}

			parent_dir = obj;

			obj = (struct ObjNode *)IEXEC->AVL_FindNode(parent_dir->content_list, (AVLKey)name, avl_key_compare_func);
			if (NO obj) {
				obj = alloc_obj_node(gd, name, parent_dir);
				if (NO obj) {
					return DOSFALSE;
				}
			}
		} while (path_pos != -1);

		obj->xad_file_info = xad_file_info;

		if (NOT (xad_file_info->xfi_Flags & XADFIF_DIRECTORY) &&
			path[IUTILITY->Strlen(path) - 1] != '/') /* workaround for lha client */
		{
			obj->type = FSO_TYPE_FILE;

			if (obj->content_list) {
				IDOS->SetIoErr(ERROR_OBJECT_WRONG_TYPE); /* only directories can have contents */
				return DOSFALSE;
			}

			if (NOT (xad_file_info->xfi_Flags & XADFIF_NOUNCRUNCHSIZE)) {
				obj->file_size = xad_file_info->xfi_Size;
				gd->disk_size += obj->file_size;
			}
		}

		IXAD->xadConvertProtection(
			XAD_PROTFILEINFO,     xad_file_info,
			XAD_GETPROTAMIGA,     &obj->protection,
			TAG_END);

		IXAD->xadConvertDates(
			XAD_DATEXADDATE,      &xad_file_info->xfi_Date,
			XAD_GETDATEDATESTAMP, &obj->date,
			TAG_END);

		xad_file_info = xad_file_info->xfi_Next;
	}

	return DOSTRUE;
}

static void free_xad_dir_tree(struct GlobalData *gd) {
	free_obj_node(gd, gd->root_node);
	gd->root_node = NULL;
	gd->disk_size = 0;
}

static struct ObjNode *alloc_obj_node(struct GlobalData *gd, CONST_STRPTR name,
	struct ObjNode *parent_dir)
{
	struct ObjNode *obj;

	obj = IEXEC->AllocPooled(gd->memory_pool, sizeof(*obj));
	if (NO obj) {
		IDOS->SetIoErr(ERROR_NO_FREE_STORE);
		return NULL;
	}
	
	IUTILITY->ClearMem(obj, sizeof(*obj));

	obj->global_data = gd;
	obj->type = FSO_TYPE_DIRECTORY;
	obj->file_size = -1;
	IDOS->DateStamp(&obj->date);

	if (name && name[0] != '\0') {
		obj->name_size = IUTILITY->Strlen(name) + 1;
		obj->name = IEXEC->AllocPooled(gd->memory_pool, obj->name_size);
		if (NO obj->name) {
			free_obj_node(gd, obj);
			IDOS->SetIoErr(ERROR_NO_FREE_STORE);
			return NULL;
		}

		IUTILITY->Strlcpy((STRPTR)obj->name, name, obj->name_size);
	}

	if (parent_dir) {
		obj->parent_dir = parent_dir;
		if (IEXEC->AVL_AddNode(&parent_dir->content_list, &obj->node, avl_node_compare_func)) {
			free_obj_node(gd, obj);
			IDOS->SetIoErr(ERROR_OBJECT_EXISTS); /* should never happen */
			return DOSFALSE;
		}
	}

	return obj;
}

static void free_obj_node(struct GlobalData *gd, struct ObjNode *obj) {
	if (obj) {
		if (obj->content_list) {
			struct AVLNode *node;
			while ((node = IEXEC->AVL_FindFirstNode(obj->content_list))) {
				IEXEC->AVL_RemNodeByAddress(&obj->content_list, node);
				free_obj_node(gd, (struct ObjNode *)node);
			}
		}
		IEXEC->FreePooled(gd->memory_pool, (STRPTR)obj->name, obj->name_size);
		IEXEC->FreePooled(gd->memory_pool, obj, sizeof(*obj));
	}
}

static int32 create_cache_directory(struct GlobalData *gd) {
	TEXT dirPath[1024];
	BPTR dirLock, oldLock;

	if (gd->cache_dir_lock) {
		return DOSTRUE;
	}

	if (IDOS->GetVar("XADFS_CACHE_DIR", dirPath, sizeof(dirPath), 0) > 0) {
		dirLock = IDOS->Lock(dirPath, SHARED_LOCK);
	} else {
		dirLock = IDOS->Lock("T:", SHARED_LOCK);
	}
	if (NO dirLock) {
		return DOSFALSE;
	}

	oldLock = IDOS->CurrentDir(dirLock);

	gd->cache_dir_lock = IDOS->Lock(gd->device_name, EXCLUSIVE_LOCK);
	if (NO gd->cache_dir_lock) {
		gd->cache_dir_lock = IDOS->CreateDir(gd->device_name);
	}

	IDOS->CurrentDir(oldLock);

	IDOS->UnLock(dirLock);

	if (gd->cache_dir_lock) {
		delete_cache_directory_contents(gd, gd->cache_dir_lock);
		return DOSTRUE;
	} else {
		return DOSFALSE;
	}
}

static void delete_cache_directory(struct GlobalData *gd) {
	if (gd->cache_dir_lock) {
		TEXT dirPath[1024];

		delete_cache_directory_contents(gd, gd->cache_dir_lock);

		IDOS->DevNameFromLock(gd->cache_dir_lock, dirPath, sizeof(dirPath), DN_FULLPATH);

		IDOS->UnLock(gd->cache_dir_lock);

		IDOS->Delete(dirPath);

		gd->cache_dir_lock = ZERO;
	}
}

static void delete_cache_directory_contents(struct GlobalData *gd, BPTR dirLock) {
	if (dirLock) {
		APTR ctx;
		ctx = IDOS->ObtainDirContextTags(
			EX_FileLockInput, dirLock,
			EX_DataFields,    EXF_NAME,
			EX_DoCurrentDir,  TRUE,
			TAG_END);
		if (ctx) {
			struct ExamineData *exd;
			while ((exd = IDOS->ExamineDir(ctx))) {
				if (IUTILITY->Strnicmp(exd->Name, "XADFS.", 6) == SAME) {
					IDOS->Delete(exd->Name);
				}
			}
			IDOS->ReleaseDirContext(ctx);
		}
	}
}

static int32 extract_xad_file(struct GlobalData *gd, int32 *res2, struct ObjNode *obj) {
	TEXT fileName[256];
	BPTR oldLock;
	BPTR cacheFile;
	int32 xad_error;

	if (obj->type != FSO_TYPE_FILE)	{
		*res2 = ERROR_OBJECT_WRONG_TYPE;
		return DOSFALSE;
	}

	if (NOT create_cache_directory(gd)) {
		*res2 = IDOS->IoErr();
		return DOSFALSE;
	}

	IUTILITY->SNPrintf(fileName, sizeof(fileName), "XADFS.%08lX.%s",
		MKBADDR(obj), obj->name);

	oldLock = IDOS->CurrentDir(gd->cache_dir_lock);

	cacheFile = IDOS->Open(fileName, MODE_NEWFILE);

	IDOS->CurrentDir(oldLock);

	if (NO cacheFile) {
		*res2 = IDOS->IoErr();
		return DOSFALSE;
	}

	xad_error = IXAD->xadFileUnArc(gd->xad_archive_info,
		XAD_OUTFILEHANDLE, cacheFile,
		XAD_ENTRYNUMBER,   obj->xad_file_info->xfi_EntryNumber,
		TAG_END);

	IDOS->Close(cacheFile);

	if (xad_error) {
		IDOS->Delete(fileName);
		*res2 = xad_to_ioerror(xad_error);
		return DOSFALSE;
	}

	return DOSTRUE;
}

BPTR get_cache_file(struct GlobalData *gd, int32 *res2, struct ObjNode *obj) {
	TEXT fileName[256];
	BPTR oldLock;
	BPTR cacheFile;
	int32 tryCount;

	if (obj->type != FSO_TYPE_FILE)	{
		*res2 = ERROR_OBJECT_WRONG_TYPE;
		return ZERO;
	}

	if (NO create_cache_directory(gd)) {
		*res2 = IDOS->IoErr();
		return ZERO;
	}

	IUTILITY->SNPrintf(fileName, sizeof(fileName), "XADFS.%08lX.%s",
		MKBADDR(obj), obj->name);

	tryCount = 0;
	do {
		oldLock = IDOS->CurrentDir(gd->cache_dir_lock);

		cacheFile = IDOS->Open(fileName, MODE_OLDFILE);

		IDOS->CurrentDir(oldLock);

		if (cacheFile) {
			return cacheFile;
		}

		if (NOT extract_xad_file(gd, res2, obj)) {
			return ZERO;
		}
	} while (++tryCount < 2);

	*res2 = IDOS->IoErr();
	return ZERO;
}

